
/**
 * This class represents the current context for printable and/or whitespace
 * characters, as layed out in the Spec (Section 3)...
 *
 * No rocket science here... 
 * 
 * Use an explicit UNDEFINED for the initial (unset) states for color and size..
 **/
public class Context {
  boolean bFlag;
  boolean emFlag;
  boolean iFlag;
  boolean sFlag;
  boolean ttFlag;
  int u;
  int size;
  int color;

  static int UNDEFINED = 999;

  public Context(Context other) {
    this.bFlag = other.bFlag;
    this.emFlag = other.emFlag;
    this.iFlag = other.iFlag;
    this.sFlag = other.sFlag;
    this.ttFlag = other.ttFlag;
    this.u = other.u;
    this.size = other.size;
    this.color = other.color;
  }

  public void copy(Context other) {
    this.bFlag = other.bFlag;
    this.emFlag = other.emFlag;
    this.iFlag = other.iFlag;
    this.sFlag = other.sFlag;
    this.ttFlag = other.ttFlag;
    this.u = other.u;
    this.size = other.size;
    this.color = other.color;
  }

  public boolean equals(Context other) {
    if (this.bFlag == other.bFlag && this.emFlag == other.emFlag
         && this.iFlag == other.iFlag && this.sFlag == other.sFlag
         && this.ttFlag == other.ttFlag &&  this.u == other.u
         && this.size == other.size && this.color == other.color) {
      return(true);
    } else {
      return(false);
    }
  }

  public boolean spaceEquals(Context other) {
    // This is a workaround to never compress spaces in TT Context...
    if (this.ttFlag || other.ttFlag) {
      return(false);
    }
    if (this.u == other.u && this.size == other.size) {
      if (this.u == 0) {
        return(true); // color doesn't matter...
      } else {
        if (this.color == other.color) {
          return(true);
        } 
      }
    } 
    return(false);
  }

  public Context() {
    bFlag = emFlag = iFlag = sFlag = ttFlag = false;
    u = 0;
    size = UNDEFINED;
    color = UNDEFINED;
  }

  public void setTag(char tag) {
    if (tag == Main.BOLD) {
      bFlag = true;
    } else if (tag == Main.EM) {
      if (!sFlag) {
        emFlag = !emFlag;
      }
    } else if (tag == Main.ITALIC) {
      iFlag = true;
    } else if (tag == Main.PLAIN) {
      u = 0; sFlag = emFlag = iFlag = bFlag = ttFlag = false;
    } else if (tag == Main.STRONG) {
      sFlag = true; 
      emFlag = false;
    } else if (tag == Main.TYPEWRITER) {
      ttFlag = true;
    } else if (tag == Main.UNDERLINE) {
      if (u < 3) { 
        u++; 
      }
    } else if (tag >= Main.ZERO && tag <= Main.NINE) {
      size = tag - Main.ZERO;
    } else if (tag >= Main.RED && tag <= Main.WHITE) {
      color = tag;
    } else if (tag == Main.SKIPPOPTAG) {
      // ignore...
    } else {
      System.err.println("UNEXPECTED TAG!");
    }
  }

  public String toString() {
    String ret = "";
    if (bFlag) {
      ret += "b;";
    }
    if (ttFlag) {
      ret += "tt;";
    }
    if (sFlag) {
      ret += "s;";
    }
    if (iFlag) {
      ret += "i;";
    }
    if (emFlag) {
      ret += "em;";
    }
    ret += "C:" + color + " S:" + size + " U: " + u;
    return(ret);
  }

  public boolean getTTFlag() {
    return(ttFlag);
  }

  public boolean getSFlag() {
    return(sFlag);
  }

}
