import java.util.*;

public class Document {
  // public?  Yes... This is evil... Just a glorified struct...
  public char mTag = 0;
  public Vector mContents = new Vector();
  public Document mParent = null;
  public int mPlaceHolder = 0;
  public Context mContext = null;

  public Document(Document parent, char tag) {
    mParent = parent;
    mTag = tag;
    mContext = new Context(parent.mContext);
    mContext.setTag(mTag);
  }

  public Document() {
    mContext = new Context();
  }

  public void append(Document doc) {
    if (this.mTag != doc.mTag) {
      mContents.addElement(doc);
      doc.mParent = this;
    } else {
      for(int c=0;c<doc.mContents.size();++c) {
        Object obj = doc.mContents.elementAt(c);
        if (obj instanceof Document) {
          Document temp = (Document)obj;
          temp.mParent = this;
        }
        mContents.addElement(obj);
      }
    }
  }

  public String toString() {
    return(toString(0));
  }

  public String toString(int indentLen) {
    StringBuffer ret = new StringBuffer("");
    StringBuffer indent = new StringBuffer("");
    for(int c=0;c<indentLen;++c) {
      indent.append(" ");
    }
    ret.append(indent + "{" + (int)mTag + "," + mPlaceHolder + "\n");
    for(int c=0;c<mContents.size();++c) {
      Object obj = mContents.elementAt(c);
      if (obj instanceof Document) {
        ret.append(((Document)obj).toString(indentLen+2));
      } else {
        ret.append("  " + indent+obj.toString() + "\n");
      }
    }
    ret.append(indent + "}\n");
    return(ret.toString());
  }

  /**
   * Check if we're empty... if we are, remove ourselves from the hierarchy...
   * update parents placeholder...
   */
  public boolean isEmpty() {
    if (mContents.size() == 0) {
      if (mParent != null) {
        // I've got nothing...
        mParent.mContents.removeElement(this);
        mParent.mPlaceHolder--;
      }
      return(true);
    }
    return(false);
  }

  public boolean reCalculateContext() {
    mContext.copy(mParent.mContext);
    mContext.setTag(mTag);
    return(true);
  }

  /**
   * Look at consecutive children for mergeable entities...
   * <b>XXX</b><b>YYY</b> -> <b>XXXYYY</b>
   * Fairly simple...
   **/
  public boolean mergeChildren() {
    boolean ret = false;
    if (mContents.size() < 2) {
      return(ret);
    }
    char oldTag = getTag(mContents.elementAt(0));
    for(int c=1;c<mContents.size();++c) {
      char tag = getTag(mContents.elementAt(c));
      if (oldTag != 0 && tag != 0) {
        Document d1 = (Document)mContents.elementAt(c-1);
        Document d2 = (Document)mContents.elementAt(c);
        if (tag == oldTag || d1.mContext.equals(d2.mContext)) {
          d1.append(d2);
          mContents.removeElement(d2);
          c--;
          ret = true;
        }
      }
      oldTag = tag;
    }
    return(ret);
  }

  private static Vector getSingleTags(Object obj) {
    Vector v = new Vector();
    if (obj instanceof Document) {
      Document d = (Document)obj;
      v.addElement(new Integer(d.mTag));
      while(d != null && d.mContents.size() == 1) {
        obj = d.mContents.firstElement();
        if (obj instanceof Document) {
          d = (Document)obj;
          v.addElement(new Integer(d.mTag));
        } else {
          d = null;
        }
      }
    }
    return(v);
  }

  /**
   * Ok...
   * We can't always switch nested tags... <b><r>XXX</r><b> != <r><b>XXX</b></r>
   * We want to check in a vector of tags, that switching the tags at i1 and
   * i2 will not impact anything...
   */
  private boolean canSwitch(Vector v, int i1, int i2) {
    if (i1 == i2) {
      return(true);
    }
    // Check for PL... let's not mess with this...
    int colorCount = 0;
    int sizeCount = 0;
    Integer pl = new Integer(Main.PLAIN);
    for(int c=i1;c<i2;++c) {
      Integer val = (Integer)v.elementAt(c);
      if (val.equals(pl)) {
        return(false);
      }
      if (isColor((char)val.intValue())) {
        colorCount++;
      }
      if (isSize((char)val.intValue())) {
        sizeCount++;
      }
    }
    if (sizeCount >= 2 || colorCount >= 2) {
      return(false);
    }
    return(true);
  }

  /**
   * See the comments within this class to see what it does...
   */
  public boolean checkOverlap() {
    boolean ret = false;

    // Our first visit to this node... Let's check it out...
    Vector cv = mContents;
    int size = cv.size();
    // If size is one, we can switch spots with out parent without affecting
    // context... also if parent is a color or size and so are we, we
    // can replace parent... 
    // code moved out of here...

    // If size is 2... then we've got enough children to look and see
    // if they can be mashed together somehow...
    // If our children are purely wrappers around other docs, we're
    // going to consider that as well (see caveats in canSwitch function)...
    // This should catch:
    // <I><B>Hi There</B></I><B>Whatever</B> ->
    //    <B><I>Hi There</I>Whatver</B>... 
    if (size >= 2) {
      // Let's build a Vector of Vectors...
      // Vectors of Vectors of tags (to represent singles beneath us...)
      Vector v = new Vector();
      for(int c=0;c<size;++c) {
        v.addElement(getSingleTags(mContents.elementAt(c)));
      }
      // Ok... getSingleTags will return an empty Vector if that child is a
      // document...
      // for every item in the list... let's look at the neighbour to the
      // right... if they contain the same tag as us, we can merge ...
      for(int c=0;c<v.size()-1;++c) {
        Vector v1 = (Vector)v.elementAt(c);
        Vector v2 = (Vector)v.elementAt(c+1);
        if (v1.size() >= 1 && v2.size() >= 1 && (v1.size()+v2.size()) >= 3) {
          // There would be nothing to figure out if the above wasn't true...

          for(int d=0;d<v1.size();++d) {
            Integer v1i = (Integer)v1.elementAt(d);
            int v2index = v2.indexOf(v1i);
            if (v2index >= 0 && canSwitch(v1,0,d) && canSwitch(v2,0,v2index)) {
              // Ok... lets get some document references...
              Document d1 = (Document)mContents.elementAt(c);
              Document d1child = d1;
              Document d2 = (Document)mContents.elementAt(c+1);
              Document d2child = d2;


              // If d >= 1, we need to flip tags with d1child...
              for(int e=0;e<d;++e) {
                d1child = (Document)(d1child.mContents.firstElement());
              }
              if (d >= 1) {
                char tempTag = d1.mTag;
                d1.mTag = d1child.mTag;
                d1child.mTag = tempTag;
              }

              // If v2index >= 1, we need to flip tags with d2child...
              for(int e=0;e<v2index;++e) {
                d2child = (Document)(d2child.mContents.firstElement());
              }
              if (v2index >= 1) {
                char tempTag = d2.mTag;
                d2.mTag = d2child.mTag;
                d2child.mTag = tempTag;
              }
              // Now let's just append d2 to d1...
              d1.append(d2);
              // And remove d1 from his parent...
              mContents.removeElement(d2);
              // Our vectors are a total mess... exit out of here...
              return(true);
            }
          }
        }
      }
    }
    if (size >= 3) {
      // Many tags at the same level... potential for:
      // <r>xxx</r><g>xxx</g><r>xxx</r>
      //  -> <r>xxx<g>xxx</g>xxx</r>
      // This won't work:
      // <r>xxx</r><B>xxx</B><r>xxx</r>
      // ... !also works for sizes!...
      Vector colors = new Vector();
      boolean colorsContainsSizes = false;
      for(int c=0;c<cv.size();++c) {
        char ch = getTag(cv.elementAt(c));
        if (isColor(ch) || isSize(ch)) {
          Integer i = new Integer(ch);
          if (colors.contains(i)) {
            // Found something... rearrange...
            int end = c;
            int start = c-colors.size()+colors.indexOf(i);
            Document keeper = (Document)cv.elementAt(start);
            for(int d=start+1;d<=end;++d) {
              // Don't use 'd' since our Vectors are shrinking!
              Document merger = (Document)cv.elementAt(start+1);
              keeper.append(merger);
              cv.removeElement(merger);
              colors.removeElementAt(colors.size()-1); // order doesn't matter
              c--;
              ret = true;
            }
          } else {
            if (colorsContainsSizes && isColor(ch)) {
              colors.clear();
              colorsContainsSizes = false;
            } else if (!colorsContainsSizes && isSize(ch)) {
              colors.clear();
              colorsContainsSizes = true;
            }
            colors.addElement(i);
          }
        } else {
          colors.clear();
        }
      }
    } // size >= 3...
    return(ret);
  }

  public static char getTag(Object obj) {
    if (obj instanceof Document) {
      Document d = (Document)obj;
      return(d.mTag);
    }
    return(0);
  }

  public static boolean isColor(char ch) {
    if (ch >= Main.RED && ch <= Main.WHITE) {
      return(true);
    }
    return(false);
  }

  public static boolean isSize(char ch) {
    if (ch >= Main.ZERO && ch <= Main.NINE) {
      return(true);
    }
    return(false);
  }


  // This function looks to see if this Document is a wrapper around
  // a single Document (and if those are single, considers them...)
  // It looks to squeeze colors and sizes down...
  // <1><2>Whatever</2></1> -> <2>Whatever</2>
  // The only way we'll get really deep documents is through using
  // colors, sizes, or the <PL> Tag (not handled here...)
  // Hmmm... <B><PL>XXX</PL></B> is the same as <PL>XXX</PL>... Hmmm... 
  public boolean checkSqueeze() {
    if (mContents.size() == 1) {
      if (isColor(mTag) || isSize(mTag)) {
        // If a color/size exists in a series of single nodes below us, we
        // can be removed...
        Object obj = mContents.firstElement();
        if (obj instanceof Document) {
          Document child = (Document)obj;
          Document immediateChild = child;  // child will move further down
                                            // list...
          while(child != null) {
            char ch = child.mTag;
            if ((isSize(ch) && isSize(mTag)) || (isColor(ch) && isColor(mTag))) {
              // our child is a colour or size... so are we... we are redundant...
              // An easier way to do this is just take our immediate child's
              // context and tag!
              mTag = immediateChild.mTag;
              mContents = immediateChild.mContents;
              mContext = immediateChild.mContext; // (don't think this matters)
              // Keep mParent... and mParent's reference to us!
              for(int c=0;c<mContents.size();++c) {
                Object content = mContents.elementAt(c);
                if (content instanceof Document) {
                  ((Document)content).mParent = this;
                }
              }
              return(true);
            } else {
              if (child.mContents.size() == 1 
                     && child.mContents.firstElement() instanceof Document) {
                child = (Document)child.mContents.firstElement();
              } else {
                child = null;
              }
            } 
          } // while child != null...
        }
      }
    }
    return(false);

  }

}
