import java.util.*;
import java.io.*;
import java.lang.reflect.Method;

/**
 * This class contains most of the driving functions for processing...
 * All input is done here (reading stdin into a preallocated buffer of 6 Meg)
 * Some processing is done during this reading phase...
 * Both algorithms clean spaces out in the same fashion...
 * Alg2 goes on to build a Document tree, looking for matching tags, etc...
 * Alg1 is simpler, and on output, looks to remove some unneeded tags...
 * 
 **/
public class Main {
  static char POPTAG=129;
  static char SKIPPOPTAG=255;
  static char ZERO=200; // ONE = -101, -102...
  static char NINE=209;
  static char TYPEWRITER=130;
  static char PLAIN=131;
  static char BOLD=132;
  static char ITALIC=133;
  static char STRONG=134;
  static char EM=135;
  static char UNDERLINE=136;
  // colours...
  static char RED=180;
  static char GREEN=181;
  static char BLUE=182;
  static char CYAN=183;
  static char MAGENTA=184;
  static char YELLOW=185;
  static char BLACK=186;
  static char WHITE=187;

  char mBuffer[];
  int mCount;
  Hashtable mTagHash = new Hashtable();
  Document mRoot = null;

/*****
 *** CODE COMMON TO ALG1 AND ALG2...
 ****/
 
  Main() {
    mBuffer = new char[6000000];
    mCount = 0;
    mTagHash.put(new Integer(RED),"r");
    mTagHash.put(new Integer(GREEN),"g");
    mTagHash.put(new Integer(BLUE),"b");
    mTagHash.put(new Integer(CYAN),"c");
    mTagHash.put(new Integer(MAGENTA),"m");
    mTagHash.put(new Integer(YELLOW),"y");
    mTagHash.put(new Integer(BLACK),"k");
    mTagHash.put(new Integer(WHITE),"w");
    for(int c=0;c<=9;++c) {
      mTagHash.put(new Integer(ZERO+c),Integer.toString(c));
    }
    mTagHash.put(new Integer(TYPEWRITER),"TT");
    mTagHash.put(new Integer(PLAIN),"PL");
    mTagHash.put(new Integer(BOLD),"B");
    mTagHash.put(new Integer(ITALIC),"I");
    mTagHash.put(new Integer(STRONG),"S");
    mTagHash.put(new Integer(EM),"EM");
    mTagHash.put(new Integer(UNDERLINE),"U");
  }

  //
  // This function will perform our space concatenation...
  // And will remove unnecessary <TAG></TAG> entries...
  // And will remove EM's in S's... By checking lack of change in context...
  //   so this also catches <b><b>Whatever</b></b>
  //
  public void inputBuffer() {
   Vector contextStack = new Vector();

   Context context = new Context();

   try {
    InputStream is = new BufferedInputStream(System.in);

    mCount = 0; // offset into buffer...

    int ch;
    while((ch = is.read()) != -1) {
      if (ch == '<') {
        // A tag...
        int nextCh = is.read();
        if (nextCh == '/') {
          int nextNextCh = is.read();
          if (mBuffer[mCount-1] >= 128 && mBuffer[mCount-1] != POPTAG) {
            // we're popping off a tag we just put on...
            mCount--;
          } else {
            mBuffer[mCount++] = POPTAG;
            // closing a tag - POP...
            // we don't need the read... keep reading...
          }
          while((ch = is.read()) != '>'); // we can do this because do is valid.

          // Pop out last context off the stack...
	  context = (Context)contextStack.lastElement();
          contextStack.removeElementAt(contextStack.size()-1);
        } else {
          char tag;
          if (nextCh >= '0' && nextCh <= '9') {
            tag = (char)(ZERO + (nextCh-'0'));
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'B') {
	    tag = BOLD;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'I') {
	    tag = ITALIC;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'S') {
	    tag = STRONG;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'U') {
	    tag = UNDERLINE;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'r') {
	    tag = RED;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'g') {
	    tag = GREEN;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'b') {
	    tag = BLUE;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'c') {
	    tag = CYAN;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'm') {
	    tag = MAGENTA;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'y') {
	    tag = YELLOW;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'k') {
	    tag = BLACK;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'w') {
	    tag = WHITE;
            mBuffer[mCount++] = tag;
          } else if (nextCh == 'E') {
            // Must be EM
            // Only add if we're not in S...
            tag = EM;
            mBuffer[mCount++] = tag;
            is.read();
          } else if (nextCh == 'P') {
	    tag = PLAIN;
            mBuffer[mCount++] = tag;
            is.read();
          } else if (nextCh == 'T') {
	    tag = TYPEWRITER;
            mBuffer[mCount++] = tag;
            is.read();
          } else {
            // TODO: Remove...
            tag = PLAIN; // NO NOT REALLY...
            System.err.println("Unknown nextCh: " + nextCh);
          }
          is.read();
          // Push our current context onto the stack...
          Context oldContext = context;
          contextStack.addElement(oldContext);
          // copy and update our context...
          context = new Context(context);
          context.setTag(tag);
          if (context.equals(oldContext)) {
            // Tag added did not affect context... Let's throw it and it's
            // buddy out...
            mBuffer[mCount-1] = SKIPPOPTAG;
          }
        }
      } else if (ch == ' ' || ch == 10 || ch == 13 || ch == 9) {
        // whitespace...
        if (mCount == 0) { 
          mBuffer[mCount++] = ' ';
        } else if (context.getTTFlag()) {
          mBuffer[mCount++] = (char)ch;
        } else if (mBuffer[mCount-1] == ' ' && !context.getTTFlag()) {
          ; 
        } else {
          mBuffer[mCount++] = ' '; 
        }
      } else {
        // must be a printable character...
        mBuffer[mCount++] = (char)ch;
      }
    }
    // finished reading...
   } catch (Exception e) {
    e.printStackTrace();
   }


  }


  /**
   * This will update out buffer array, removing spaces
   * that are unnecessary...
   * Updates mBuffer and mCount...
   **/
  public void removeSpacesConsideringContext() {
    Vector contextStack = new Vector();
    Context currentContext = new Context();
    boolean lastPrintableCharSpace = false;
    Context lastSpaceContext = null;
    int dest = 0;
    char ch;
    for(int c=0;c<mCount;++c) {
      ch = mBuffer[c];
      if (ch == ' ') {
        if (lastPrintableCharSpace) {
          // We just printed a space... Compare contexts...
          if (lastSpaceContext.spaceEquals(currentContext)) {
            // Contexts match... skip this char... (don't update dest!)
          } else {
            // Don't match... print it...
            // update lastSpaceContext...
            mBuffer[dest++] = ch;
            lastSpaceContext = new Context(currentContext);
          }
        } else {
          lastPrintableCharSpace = true;
          lastSpaceContext = new Context(currentContext);
          // We want this one added...
          mBuffer[dest++] = ch;
        }
      } else {
        // We don't care... Copy across...
        mBuffer[dest++] = ch;

        // But if it's a tag... We need to update out context(s)...
        if (ch == POPTAG) {
          // Pop out last context off the stack...
	  currentContext = (Context)contextStack.lastElement();
          contextStack.removeElementAt(contextStack.size()-1);
        } else if (ch > 128) {
          // Push out context onto the stack, and update our current context...
          Context oldContext = currentContext;
          contextStack.addElement(oldContext);
          // copy and update our context...
          currentContext = new Context(currentContext);
          currentContext.setTag(ch);
        } else {
          // This was a printable character (and not a space!)
          lastPrintableCharSpace=false;
        }
      }
    }
    mCount = dest;
  }

/******
 *** ALG2 ONLY...
 ****/

  /**
   * Ok... We're going to build a document tree... Recursion won't work 100K
   * items deep (which could happen with a 5Meg doc...)
   **/
  public void buildDocumentTree() {
    mRoot = new Document();
    Document currentDocument = mRoot;
    Document parent;
    StringBuffer sb = new StringBuffer();
    for(int c=0;c<mCount;++c) {
      char ch = mBuffer[c];
      if (ch < 128) {
        // printable...
        if (ch != 0) {
          sb.append(ch);
        }
      } else if (ch != POPTAG) {
        // Must be a new tag!
        if (sb.length() > 0) {
          currentDocument.mContents.addElement(sb);
          sb = new StringBuffer();
        }
        parent = currentDocument;
        currentDocument = new Document(parent,ch);
        parent.mContents.addElement(currentDocument);
      } else if (ch == POPTAG) {
        // End of our tag...
        if (sb.length() > 0) {
          currentDocument.mContents.addElement(sb);
          sb = new StringBuffer();
        }
        currentDocument = currentDocument.mParent;
      }
    }
    if (sb.length() > 0) {
      currentDocument.mContents.addElement(sb);
      sb = new StringBuffer();
    }
  }

  /**
   * remove empty documents, merge matching documents, etc...
   * <b>XXXX</b><b>xxx</b> becomes: <b>xxxxxx</b>
   * remove nodes that do not affect Context...
   **/
  public boolean pruneTree() {
    boolean ret = false;
    Document currentDocument = mRoot;

    while(currentDocument.mergeChildren()) {
      ret = true;
    }

    currentDocument.mPlaceHolder = 0;
    while(currentDocument != null) {
      if (currentDocument.mPlaceHolder < currentDocument.mContents.size()) {
        Object obj = currentDocument.mContents.elementAt(currentDocument.mPlaceHolder++);
        if (obj instanceof StringBuffer) {
          // obj is stringBuffer...
        } else {
          // obj is a Document...
          Document childDocument = (Document)obj;
          if (!childDocument.isEmpty()) {

            childDocument.reCalculateContext();
            while(childDocument.mergeChildren()) {
              ret = true;
            }

            // Child matters... work our way through...
            currentDocument = (Document)childDocument;
            currentDocument.mPlaceHolder = 0;
          }
        }
      } else {
        // The end of this document...
        currentDocument = currentDocument.mParent;
        while(currentDocument != null && currentDocument.isEmpty()) {
          currentDocument = currentDocument.mParent;
        }
      }
    }
    return(ret);
  }

  // Let's try and collapse it:
  // <4><5>whatever</5></4>
  //   becomes: <5>whatever</5>
  // <r>xxx</r><g>xxx</g><r>xxx</r> becomes <r>xxx<g>xxx</g>xxx</r>
  public boolean squeezeAll() {
    boolean ret = false;

    try {
      // We want to call checkSqueeze on our whole tree...
      Class cls = mRoot.getClass();
      Method m = cls.getMethod("checkSqueeze",null);
      if (cleanTree(m)) {
        ret = true;
      }
      m = cls.getMethod("checkOverlap",null);
      if (cleanTree(m)) {
        ret = true;
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
    return(ret);
  }

  // Ok we've built out tree... Navigation time... ideally this would be
  // be easier with recursion... BUT the tree might be too deep...
  public boolean cleanTree(Method m) throws java.lang.Exception {
    boolean ret = false;
//    System.err.println("CLEANTREE: " + mRoot);

    Document currentDocument = mRoot;
    currentDocument.mPlaceHolder = 0;
    while(m.invoke(currentDocument,null).equals(Boolean.TRUE)) {
      ret = true;
    }
    while(currentDocument != null) {
      if (currentDocument.mPlaceHolder < currentDocument.mContents.size()) {
        Object obj = currentDocument.mContents.elementAt(currentDocument.mPlaceHolder++);
        if (obj instanceof Document) {
          currentDocument = (Document)obj;
          currentDocument.mPlaceHolder = 0;

          while(m.invoke(currentDocument,null).equals(Boolean.TRUE)) {
            ret = true;
          }
        }
      } else {
        currentDocument = currentDocument.mParent;
      }
    }
    return(ret);
  }


  public void dumpTree() {
//    System.out.println("DUMPTREE:\n" + mRoot);
    // Ok...
    // Now lets dump out document out...
    OutputStream os = new BufferedOutputStream(System.out);

    Document currentDocument = mRoot;
    currentDocument.mPlaceHolder = 0;
    Integer i;
   try {
    while(currentDocument != null) {
      if (currentDocument.mPlaceHolder < currentDocument.mContents.size()) {
        Object obj = currentDocument.mContents.elementAt(currentDocument.mPlaceHolder++);
        if (obj instanceof StringBuffer) {
          os.write(obj.toString().getBytes());
        } else {
          // obj is a Document...
          currentDocument = (Document)obj;
          i = new Integer(currentDocument.mTag);
          if (currentDocument.mTag != SKIPPOPTAG) {
            os.write(("<" + mTagHash.get(i) + ">").getBytes());
          }
          currentDocument.mPlaceHolder = 0;
        }
      } else {
        // The end of this document...
        i = new Integer(currentDocument.mTag);
        if (currentDocument.mTag != SKIPPOPTAG) {
          if (mTagHash.get(i) != null) {
            os.write(("</" + mTagHash.get(i) + ">").getBytes());
          }
        }
        currentDocument = currentDocument.mParent;
      }
    }
    os.close();
   } catch (Exception e) {
    e.printStackTrace();
   }
    
  }


/*****
 *** ALG1 ONLY...
 ****/

  public void outputBuffer() {
   try {
    Vector tagStack = new Vector();
    Integer i = new Integer(0);
    char ch;
    OutputStream os = new BufferedOutputStream(System.out);
       
    for(int c=0;c<mCount;++c) {
      ch = mBuffer[c];
      if (ch < 128) {
        os.write(ch);
      } else {
        // it had better be a tag...
        if (ch == POPTAG) {
          i = (Integer)tagStack.lastElement();
          // Changed logic to only pull tags off the stack if </b><b> does
          // not occur...

          if (i.intValue() != SKIPPOPTAG) {
            // Check if we're about to close a tag, that we're then about
            // to open?
            if ((c+1 < mCount) && (i.intValue() == mBuffer[c+1])) {
              // Yes... increment c by one...
              c++;
            } else {
              tagStack.removeElementAt(tagStack.size()-1);
              os.write(("</" + mTagHash.get(i) + ">").getBytes());
            }
          } else {
            tagStack.removeElementAt(tagStack.size()-1);
          }
        } else {
          i = new Integer(ch);
          if (ch != SKIPPOPTAG) {
            os.write(("<" + mTagHash.get(i) + ">").getBytes());
          }
          tagStack.addElement(i);
        }
      }
    }
    os.close();
   } catch (Exception e) {
    e.printStackTrace();
   }
  }


}
